import * as i0 from '@angular/core';
import { Injectable, EventEmitter, Component, ChangeDetectionStrategy, Self, Optional, Inject, HostBinding, Input, Output, ViewChild, Directive, NgModule, InjectionToken, ChangeDetectorRef, SkipSelf, ContentChildren, ContentChild, Pipe, ViewContainerRef, TemplateRef } from '@angular/core';
import { DOCUMENT } from '@angular/common';
import { Subject, tap, withLatestFrom, skip, isObservable, filter, combineLatest, mapTo, fromEvent, takeUntil, observeOn, asapScheduler, ReplaySubject, BehaviorSubject, take, forkJoin, of, defer, catchError, throwError, map, shareReplay } from 'rxjs';
import * as THREE from 'three';
import { Clock } from 'three';
import { __rest } from 'tslib';
import { ComponentStore } from '@ngrx/component-store';

class DestroyedService extends Subject {
    ngOnDestroy() {
        this.next();
        this.complete();
    }
}
DestroyedService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: DestroyedService, deps: null, target: i0.ɵɵFactoryTarget.Injectable });
DestroyedService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: DestroyedService });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: DestroyedService, decorators: [{
            type: Injectable
        }] });

/**
 * Generate a random Id or an Event's uuid
 *
 * @param {NgtIntersection} event
 *
 * @internal
 * @private
 */
function makeId(event) {
    if (event) {
        return ((event.eventObject || event.object).uuid +
            '/' +
            event.index +
            event.instanceId);
    }
    return id();
}
const cache = {};
/**
 * Generates a short id.
 *
 * Description:
 *  A 5-character alphanumeric sequence (364 = 1.6 million)
 *  This should only be used for JavaScript specific models.
 *  http://stackoverflow.com/questions/6248666/how-to-generate-short-uid-like-ax4j9z-in-js
 *
 *  Example: `aebgf`
 */
function id() {
    let newId = ('0000' + ((Math.random() * Math.pow(36, 4)) << 0).toString(36)).slice(-4);
    newId = `a${newId}`;
    // ensure not already used
    if (!cache[newId]) {
        cache[newId] = true;
        return newId;
    }
    return id();
}

function getSelectors(store) {
    let state;
    try {
        state = store.get();
    }
    catch (e) {
        throw new Error('ComponentStore is initialized lazily.');
    }
    return Object.keys(state).reduce((selectors, key) => {
        selectors[`${key}$`] = store.select((s) => s[key]);
        return selectors;
    }, {});
}
class EnhancedComponentStore extends ComponentStore {
    constructor() {
        super(...arguments);
        this.selectors = getSelectors(this);
    }
    getImperativeState() {
        return this.get();
    }
}

class AnimationStore extends EnhancedComponentStore {
    constructor() {
        super({
            animations: {},
            objectSubscriptions: [],
            animationCallbacks: [],
            hasPriority: false,
        });
        this.animationsChangedEffect = this.effect((animations$) => animations$.pipe(tap((animations) => {
            const animationCallbacks = Object.values(animations);
            const hasPriority = animationCallbacks.some(({ priority }) => !!priority);
            this.patchState({ animationCallbacks, hasPriority });
        })));
        this.unregisterAnimationEffect = this.effect((uuid$) => uuid$.pipe(withLatestFrom(this.selectors.animations$, this.selectors.objectSubscriptions$), tap(([uuid, animations, objectSubscriptions]) => {
            const subscription = objectSubscriptions.find(([objectUuid]) => objectUuid === uuid);
            if (subscription) {
                subscription[1].unsubscribe();
            }
            const _a = animations, _b = uuid, _ = _a[_b], updatedAnimations = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            this.patchState({
                animations: updatedAnimations,
                objectSubscriptions: objectSubscriptions.filter(([objectUuid]) => objectUuid !== uuid),
            });
        })));
        this.clearEffect = this.effect((trigger$) => trigger$.pipe(tap(() => {
            this.patchState((state) => {
                state.objectSubscriptions.forEach(([, subscription]) => subscription.unsubscribe());
                return { animations: {}, objectSubscriptions: [] };
            });
        })));
        this.animationsChangedEffect(this.selectors.animations$.pipe(skip(1)));
    }
    registerAnimation(objOrObsOrCallback, callbackOrPriority, priority = 0) {
        if (objOrObsOrCallback === undefined)
            return;
        if (typeof objOrObsOrCallback === 'function') {
            const id = makeId();
            this.patchState((state) => (Object.assign(Object.assign({}, state), { animations: Object.assign(Object.assign({}, state.animations), { [id]: {
                        obj: null,
                        callback: objOrObsOrCallback,
                        priority: callbackOrPriority || 0,
                    } }) })));
            return () => {
                this.unregisterAnimationEffect(id);
            };
        }
        if (isObservable(objOrObsOrCallback)) {
            let id = '';
            const subscription = objOrObsOrCallback
                .pipe(filter((obj) => !!obj))
                .subscribe((obj) => {
                this.patchState((state) => {
                    var _a;
                    const objectSubscriptions = state.objectSubscriptions;
                    if ((_a = objectSubscriptions[objectSubscriptions.length - 1]) === null || _a === void 0 ? void 0 : _a[0]) {
                        objectSubscriptions[objectSubscriptions.length - 1][0] = obj.uuid;
                    }
                    id = obj.uuid;
                    return {
                        animations: Object.assign(Object.assign({}, state.animations), { [id]: {
                                obj,
                                callback: callbackOrPriority,
                                priority,
                            } }),
                        objectSubscriptions,
                    };
                });
            });
            this.patchState((state) => ({
                objectSubscriptions: [
                    ...state.objectSubscriptions,
                    [null, subscription],
                ],
            }));
            return () => {
                this.unregisterAnimationEffect(id);
            };
        }
        const uuid = objOrObsOrCallback.uuid;
        this.patchState((state) => ({
            animations: Object.assign(Object.assign({}, state.animations), { [uuid]: {
                    obj: objOrObsOrCallback,
                    callback: callbackOrPriority,
                    priority,
                } }),
        }));
        return () => {
            this.unregisterAnimationEffect(uuid);
        };
    }
    ngOnDestroy() {
        this.clearEffect();
        super.ngOnDestroy();
    }
}
AnimationStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: AnimationStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
AnimationStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: AnimationStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: AnimationStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return []; } });

function applyDottedPathProps(instance, key, propAtKey, isRoot = true) {
    const [first, ...paths] = key.split('.');
    const rootChild = instance[first];
    if (rootChild == null) {
        return;
    }
    if (paths.length >= 1) {
        applyDottedPathProps(rootChild, paths.join('.'), propAtKey, false);
    }
    else {
        instance[first] = propAtKey;
    }
    if (isRoot) {
        checkNeedsUpdate(rootChild);
    }
}
function checkNeedsUpdate(value) {
    if (typeof value === 'object' && 'needsUpdate' in value) {
        value['needsUpdate'] = true;
    }
}
/**
 * Apply props on instances
 *
 * @internal
 * @private
 */
function applyProps(instance, props) {
    if (!props || (props && !Object.keys(props).length))
        return;
    if ('__ngt' in props) {
        instance['__ngt'] = props['__ngt'];
    }
    const unknownInstance = instance;
    if (unknownInstance['set'] != null &&
        typeof unknownInstance['set'] === 'function') {
        unknownInstance['set'](props);
    }
    for (const [key, propAtKey] of Object.entries(props)) {
        if (key.split('.').length > 1) {
            applyDottedPathProps(instance, key, propAtKey);
        }
        else {
            const threeInstancePropAtKey = unknownInstance[key];
            if (threeInstancePropAtKey == null) {
                unknownInstance[key] = propAtKey;
            }
            else {
                if (threeInstancePropAtKey['set'] != null &&
                    typeof threeInstancePropAtKey['set'] === 'function') {
                    if (Array.isArray(propAtKey)) {
                        if (threeInstancePropAtKey['fromArray'] != null &&
                            typeof threeInstancePropAtKey['fromArray'] === 'function') {
                            threeInstancePropAtKey['fromArray'](propAtKey);
                        }
                        else {
                            threeInstancePropAtKey['set'](...propAtKey);
                        }
                    }
                    else {
                        threeInstancePropAtKey['set'](propAtKey);
                    }
                }
                else {
                    unknownInstance[key] = propAtKey;
                }
            }
            checkNeedsUpdate(propAtKey);
        }
    }
}

const position = new THREE.Vector3();
const defaultTarget = new THREE.Vector3();
const isOrthographicCamera = (value) => value && value.isOrthographicCamera;
class CanvasStore extends EnhancedComponentStore {
    constructor() {
        super({
            isOrthographic: false,
            isLinear: false,
            shadows: false,
            alpha: true,
            mouse: new THREE.Vector2(),
            clock: new THREE.Clock(),
            renderer: undefined,
            scene: undefined,
            camera: undefined,
            raycaster: undefined,
            internal: {
                active: false,
                size: { width: 0, height: 0 },
                dpr: 1,
                viewport: {
                    initialDpr: 1,
                    dpr: 1,
                    width: 0,
                    height: 0,
                    aspect: 0,
                    distance: 0,
                    factor: 0,
                    getCurrentViewport: (camera = this.getImperativeState().camera, target = defaultTarget, size = this.getImperativeState().internal.size) => {
                        const { width, height } = size;
                        const aspect = width / height;
                        const distance = camera
                            .getWorldPosition(position)
                            .distanceTo(target);
                        if (isOrthographicCamera(camera)) {
                            return {
                                width: width / camera.zoom,
                                height: height / camera.zoom,
                                factor: 1,
                                distance,
                                aspect,
                            };
                        }
                        const fov = (camera.fov * Math.PI) / 180; // convert vertical fov to radians
                        const h = 2 * Math.tan(fov / 2) * distance; // visible height
                        const w = h * (width / height);
                        return {
                            width: w,
                            height: h,
                            factor: width / w,
                            distance,
                            aspect,
                        };
                    },
                },
            },
        });
        this.setIsOrthographic = this.updater((state, isOrthographic) => (Object.assign(Object.assign({}, state), { isOrthographic })));
        this.setIsLinear = this.updater((state, isLinear) => (Object.assign(Object.assign({}, state), { isLinear })));
        this.setShadows = this.updater((state, shadows) => (Object.assign(Object.assign({}, state), { shadows })));
        this.setAlpha = this.updater((state, alpha) => (Object.assign(Object.assign({}, state), { alpha })));
        this.setSize = this.updater((state, size) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { size }) })));
        this.setDpr = this.updater((state, dpr) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { dpr }) })));
        this.setViewport = this.updater((state) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { viewport: Object.assign(Object.assign(Object.assign({}, state.internal.viewport), state.internal.viewport.getCurrentViewport(state.camera, defaultTarget, state.internal.size)), { dpr: state.internal.dpr }) }) })));
        this.setActive = this.updater((state, active) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { active }) })));
        this.initRendererEffect = this.effect((canvas$) => {
            return canvas$.pipe(withLatestFrom(this.selectors.internal$, this.selectors.isLinear$, this.selectors.shadows$, this.selectors.alpha$), tap(([canvas, { size, dpr }, isLinear, shadows, alpha]) => {
                // TODO: custom renderer
                const renderer = new THREE.WebGLRenderer({
                    canvas,
                    antialias: true,
                    powerPreference: 'high-performance',
                    alpha,
                });
                if (shadows) {
                    renderer.shadowMap.enabled = true;
                    if (typeof shadows === 'object')
                        Object.assign(renderer.shadowMap, shadows);
                    else
                        renderer.shadowMap.type = THREE.PCFSoftShadowMap;
                }
                if (!isLinear) {
                    renderer.toneMapping = THREE.ACESFilmicToneMapping;
                    renderer.outputEncoding = THREE.sRGBEncoding;
                }
                renderer.setClearAlpha(0);
                renderer.setPixelRatio(dpr);
                renderer.setSize(size.width, size.height);
                this.patchState({ renderer });
            }));
        });
        this.initSceneEffect = this.effect((sceneOptions$) => sceneOptions$.pipe(tap((sceneOptions) => {
            const scene = new THREE.Scene();
            applyProps(scene, sceneOptions);
            this.patchState({ scene });
        })));
        this.initCameraEffect = this.effect((cameraOptions$) => cameraOptions$.pipe(withLatestFrom(this.selectors.isOrthographic$, this.selectors.internal$), tap(([cameraOptions, isOrthographic, { size }]) => {
            var _a;
            const isCamera = cameraOptions instanceof THREE.Camera;
            let camera;
            if (isCamera) {
                camera = cameraOptions;
            }
            else {
                if (isOrthographic) {
                    camera = new THREE.OrthographicCamera(0, 0, 0, 0, 0.1, 1000);
                    camera.zoom = 100;
                }
                else {
                    camera = new THREE.PerspectiveCamera(75, (_a = size.width / size.height) !== null && _a !== void 0 ? _a : 0, 0.1, 1000);
                }
                camera.position.z = 5;
                applyProps(camera, cameraOptions);
                // Update projection matrix after applying props
                camera.updateProjectionMatrix();
            }
            // look at center by default
            camera.lookAt(0, 0, 0);
            this.patchState({ camera });
            this.setViewport();
        })));
        this.initRaycasterEffect = this.effect((raycasterOptions$) => raycasterOptions$.pipe(tap((raycasterOptions) => {
            const raycaster = new THREE.Raycaster();
            raycaster.enabled = true;
            applyProps(raycaster, raycasterOptions);
            this.patchState({ raycaster });
        })));
        this.windowResizeEffect = this.effect((params$) => params$.pipe(withLatestFrom(this.selectors.renderer$, this.selectors.camera$), tap(([{ dpr, size }, renderer, camera]) => {
            if (renderer && camera) {
                if (camera.type === 'PerspectiveCamera') {
                    camera.aspect = size.width / size.height;
                }
                else {
                    camera.left = size.width / -2;
                    camera.right = size.width / 2;
                    camera.top = size.height / 2;
                    camera.bottom = size.height / -2;
                }
                camera.updateProjectionMatrix();
                camera.updateMatrixWorld();
                renderer.setPixelRatio(dpr);
                renderer.setSize(size.width, size.height);
                this.setSize(size);
                this.setDpr(dpr);
                this.setViewport();
            }
        })));
        this.cleanUpEffect = this.effect(($) => $.pipe(withLatestFrom(this.selectors.renderer$), tap(([, renderer]) => {
            if (renderer) {
                renderer.renderLists.dispose();
                renderer.forceContextLoss();
            }
        })));
        this.setActive(combineLatest([
            this.selectors.renderer$,
            this.selectors.camera$,
            this.selectors.scene$,
            this.selectors.raycaster$,
        ]).pipe(filter((coreObjects) => coreObjects.every(Boolean)), mapTo(true)));
    }
    ngOnDestroy() {
        this.cleanUpEffect();
        super.ngOnDestroy();
    }
}
CanvasStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: CanvasStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
CanvasStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: CanvasStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: CanvasStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return []; } });

class LoopService {
    // TODO: rethink limit 60fps
    // private readonly interval = 1 / 60; // 60fps;
    // private internalDelta = 0;
    constructor(canvasStore, animationStore) {
        this.canvasStore = canvasStore;
        this.animationStore = animationStore;
        this.loopClock = new Clock();
    }
    start() {
        this.stop();
        const { renderer } = this.canvasStore.getImperativeState();
        if (renderer) {
            renderer.setAnimationLoop(() => {
                this.tick(this.loopClock.getDelta());
            });
        }
    }
    stop() {
        const { renderer } = this.canvasStore.getImperativeState();
        if (renderer) {
            renderer.setAnimationLoop(null);
        }
    }
    tick(delta) {
        // this.internalDelta += delta;
        //
        // if (this.internalDelta >= this.interval) {
        //   const {
        //     renderer,
        //     scene,
        //     camera,
        //     internal: { size, viewport },
        //     clock,
        //     mouse,
        //   } = this.canvasStore.getImperativeState();
        //   const { animationCallbacks, hasPriority } =
        //     this.animationStore.getImperativeState();
        //
        //   if (renderer && scene && camera) {
        //     if (hasPriority) {
        //       animationCallbacks.sort(
        //         ({ priority: a = 0 }, { priority: b = 0 }) => a - b
        //       );
        //     } else {
        //       renderer.render(scene, camera);
        //     }
        //
        //     const renderState = {
        //       clock,
        //       camera,
        //       scene,
        //       renderer,
        //       mouse,
        //       size,
        //       viewport,
        //       delta,
        //     };
        //     for (const animationCallback of animationCallbacks) {
        //       if (animationCallback.obj) {
        //         animationCallback.callback(animationCallback.obj, renderState);
        //       } else {
        //         animationCallback.callback(renderState);
        //       }
        //     }
        //   }
        //
        //   this.internalDelta = this.internalDelta % this.interval;
        // }
        const { renderer, scene, camera, internal: { size, viewport }, clock, mouse, } = this.canvasStore.getImperativeState();
        const { animationCallbacks, hasPriority } = this.animationStore.getImperativeState();
        if (renderer && scene && camera) {
            if (hasPriority) {
                animationCallbacks.sort(({ priority: a = 0 }, { priority: b = 0 }) => a - b);
            }
            else {
                renderer.render(scene, camera);
            }
            const renderState = {
                clock,
                camera,
                scene,
                renderer,
                mouse,
                size,
                viewport,
                delta,
            };
            for (const animationCallback of animationCallbacks) {
                if (animationCallback.obj) {
                    animationCallback.callback(animationCallback.obj, renderState);
                }
                else {
                    animationCallback.callback(renderState);
                }
            }
        }
    }
    ngOnDestroy() {
        this.stop();
    }
}
LoopService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoopService, deps: [{ token: CanvasStore }, { token: AnimationStore }], target: i0.ɵɵFactoryTarget.Injectable });
LoopService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoopService });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoopService, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: CanvasStore }, { type: AnimationStore }]; } });

/**
 * From r3f https://github.com/pmndrs/react-three-fiber/blob/master/packages/fiber/src/core/events.ts
 */
/** Release pointer captures.
 * This is called by releasePointerCapture in the API, and when an object is removed.
 */
function releaseInternalPointerCapture(capturedMap, obj, captures, pointerId) {
    const captureData = captures.get(obj);
    if (captureData) {
        captures.delete(obj);
        // If this was the last capturing object for this pointer
        if (captures.size === 0) {
            capturedMap.delete(pointerId);
            captureData.target.releasePointerCapture(pointerId);
        }
    }
}
function removeInteractivity(eventsStateGetter, object) {
    const { internal } = eventsStateGetter();
    // Removes every trace of an object from the data store
    internal.interaction = internal.interaction.filter((o) => o !== object);
    internal.initialHits = internal.initialHits.filter((o) => o !== object);
    internal.hovered.forEach((value, key) => {
        if (value.eventObject === object || value.object === object) {
            internal.hovered.delete(key);
        }
    });
    internal.capturedMap.forEach((captures, pointerId) => {
        releaseInternalPointerCapture(internal.capturedMap, object, captures, pointerId);
    });
}
function getRenderState(canvasState) {
    const { mouse, camera, clock, renderer, scene, internal: { size, viewport }, } = canvasState;
    const delta = clock.getDelta();
    return {
        size,
        mouse,
        camera: camera,
        delta,
        clock,
        renderer: renderer,
        scene: scene,
        viewport,
    };
}
function createEvents(canvasStateGetter, eventsStateGetter) {
    const temp = new THREE.Vector3();
    /** Sets up defaultRaycaster */
    function prepareRay(event) {
        var _a, _b;
        const state = canvasStateGetter();
        const { raycaster, mouse, camera, internal: { size }, } = state;
        // https://github.com/pmndrs/react-three-fiber/pull/782
        // Events trigger outside of canvas when moved
        if (raycaster && camera) {
            const { offsetX, offsetY } = (_b = (_a = raycaster === null || raycaster === void 0 ? void 0 : raycaster.computeOffsets) === null || _a === void 0 ? void 0 : _a.call(raycaster, event, getRenderState(state))) !== null && _b !== void 0 ? _b : event;
            const { width, height } = size;
            mouse.set((offsetX / width) * 2 - 1, -(offsetY / height) * 2 + 1);
            raycaster.setFromCamera(mouse, camera);
        }
    }
    /** Calculates delta */
    function calculateDistance(event) {
        const { internal } = eventsStateGetter();
        const dx = event.offsetX - internal.initialClick[0];
        const dy = event.offsetY - internal.initialClick[1];
        return Math.round(Math.sqrt(dx * dx + dy * dy));
    }
    /** Returns true if an instance has a valid pointer-event registered, this excludes scroll, clicks etc */
    function filterPointerEvents(objects) {
        return objects.filter((obj) => ['move', 'over', 'enter', 'out', 'leave'].some((name) => {
            var _a, _b;
            return (_b = (_a = obj.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b[('pointer' + name)];
        }));
    }
    function intersect(filter) {
        var _a;
        const canvasState = canvasStateGetter();
        const { internal } = eventsStateGetter();
        const { raycaster } = canvasState;
        // Skip event handling when noEvents is set
        if (!(raycaster === null || raycaster === void 0 ? void 0 : raycaster.enabled))
            return [];
        const seen = new Set();
        const intersections = [];
        // Allow callers to eliminate event objects
        const eventsObjects = filter
            ? filter(internal.interaction)
            : internal.interaction;
        // Intersect known handler objects and filter against duplicates
        let intersects = raycaster
            .intersectObjects(eventsObjects, true)
            .filter((item) => {
            const id = makeId(item);
            if (seen.has(id))
                return false;
            seen.add(id);
            return true;
        });
        // https://github.com/mrdoob/three.js/issues/16031
        // Allow custom userland intersect sort order
        if (raycaster.filter)
            intersects = raycaster.filter(intersects, getRenderState(canvasState));
        for (const intersect of intersects) {
            let eventObject = intersect.object;
            // Bubble event up
            while (eventObject) {
                if ((_a = eventObject.__ngt) === null || _a === void 0 ? void 0 : _a.eventCount)
                    intersections.push(Object.assign(Object.assign({}, intersect), { eventObject }));
                eventObject = eventObject.parent;
            }
        }
        return intersections;
    }
    /**  Creates filtered intersects and returns an array of positive hits */
    function patchIntersects(intersections, event) {
        const { internal } = eventsStateGetter();
        // If the interaction is captured, make all capturing targets  part of the
        // intersect.
        if ('pointerId' in event && internal.capturedMap.has(event.pointerId)) {
            for (let captureData of internal.capturedMap
                .get(event.pointerId)
                .values()) {
                intersections.push(captureData.intersection);
            }
        }
        return intersections;
    }
    function cancelPointer(hits) {
        const { internal } = eventsStateGetter();
        Array.from(internal.hovered.values()).forEach((hoveredObj) => {
            var _a, _b;
            // When no objects were hit or the the hovered object wasn't found underneath the cursor
            // we call onPointerOut and delete the object from the hovered-elements map
            if (!hits.length ||
                !hits.find((hit) => hit.object === hoveredObj.object &&
                    hit.index === hoveredObj.index &&
                    hit.instanceId === hoveredObj.instanceId)) {
                const eventObject = hoveredObj.eventObject;
                const instance = eventObject.__ngt;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                internal.hovered.delete(makeId(hoveredObj));
                if (instance === null || instance === void 0 ? void 0 : instance.eventCount) {
                    // Clear out intersects, they are outdated by now
                    const data = Object.assign(Object.assign({}, hoveredObj), { intersections: hits || [] });
                    (_a = handlers === null || handlers === void 0 ? void 0 : handlers.pointerout) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                    (_b = handlers === null || handlers === void 0 ? void 0 : handlers.pointerleave) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                }
            }
        });
    }
    /**  Handles intersections by forwarding them to handlers */
    function handleIntersects(intersections, event, delta, callback) {
        const { raycaster, mouse, camera } = canvasStateGetter();
        const { internal } = eventsStateGetter();
        // If anything has been found, forward it to the event listeners
        if (intersections.length && camera) {
            const unprojectedPoint = temp.set(mouse.x, mouse.y, 0).unproject(camera);
            const localState = { stopped: false };
            for (const hit of intersections) {
                const hasPointerCapture = (id) => { var _a, _b; return (_b = (_a = internal.capturedMap.get(id)) === null || _a === void 0 ? void 0 : _a.has(hit.eventObject)) !== null && _b !== void 0 ? _b : false; };
                const setPointerCapture = (id) => {
                    const captureData = {
                        intersection: hit,
                        target: event.target,
                    };
                    if (internal.capturedMap.has(id)) {
                        // if the pointerId was previously captured, we add the hit to the
                        // event capturedMap.
                        internal.capturedMap.get(id).set(hit.eventObject, captureData);
                    }
                    else {
                        // if the pointerId was not previously captured, we create a map
                        // containing the hitObject, and the hit. hitObject is used for
                        // faster access.
                        internal.capturedMap.set(id, new Map([[hit.eventObject, captureData]]));
                    }
                    // Call the original event now
                    event.target.setPointerCapture(id);
                };
                const releasePointerCapture = (id) => {
                    const captures = internal.capturedMap.get(id);
                    if (captures) {
                        releaseInternalPointerCapture(internal.capturedMap, hit.eventObject, captures, id);
                    }
                };
                // Add native event props
                let extractEventProps = {};
                // This iterates over the event's properties including the inherited ones. Native PointerEvents have most of their props as getters which are inherited, but polyfilled PointerEvents have them all as their own properties (i.e. not inherited). We can't use Object.keys() or Object.entries() as they only return "own" properties; nor Object.getPrototypeOf(event) as that *doesn't* return "own" properties, only inherited ones.
                for (let prop in event) {
                    let property = event[prop];
                    // Only copy over atomics, leave functions alone as these should be
                    // called as event.nativeEvent.fn()
                    if (typeof property !== 'function')
                        extractEventProps[prop] = property;
                }
                let raycastEvent = Object.assign(Object.assign(Object.assign({}, hit), extractEventProps), { spaceX: mouse.x, spaceY: mouse.y, intersections, stopped: localState.stopped, delta,
                    unprojectedPoint, ray: raycaster === null || raycaster === void 0 ? void 0 : raycaster.ray, camera: camera, 
                    // Hijack stopPropagation, which just sets a flag
                    stopPropagation: () => {
                        // https://github.com/pmndrs/react-three-fiber/issues/596
                        // Events are not allowed to stop propagation if the pointer has been captured
                        const capturesForPointer = 'pointerId' in event && internal.capturedMap.get(event.pointerId);
                        // We only authorize stopPropagation...
                        if (
                        // ...if this pointer hasn't been captured
                        !capturesForPointer ||
                            // ... or if the hit object is capturing the pointer
                            capturesForPointer.has(hit.eventObject)) {
                            raycastEvent.stopped = localState.stopped = true;
                            // Propagation is stopped, remove all other hover records
                            // An event handler is only allowed to flush other handlers if it is hovered itself
                            if (internal.hovered.size &&
                                Array.from(internal.hovered.values()).find((i) => i.eventObject === hit.eventObject)) {
                                // Objects cannot flush out higher up objects that have already caught the event
                                const higher = intersections.slice(0, intersections.indexOf(hit));
                                cancelPointer([...higher, hit]);
                            }
                        }
                    }, 
                    // there should be a distinction between target and currentTarget
                    target: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, currentTarget: {
                        hasPointerCapture,
                        setPointerCapture,
                        releasePointerCapture,
                    }, sourceEvent: event, nativeEvent: event });
                // Call subscribers
                callback(raycastEvent);
                // Event bubbling may be interrupted by stopPropagation
                if (localState.stopped)
                    break;
            }
        }
        return intersections;
    }
    const handlePointer = (name) => {
        // Deal with cancelation
        switch (name) {
            case 'pointerleave':
            case 'pointercancel':
                return () => cancelPointer([]);
            case 'onLostPointerCapture':
                return (event) => {
                    if ('pointerId' in event) {
                        // If the object event interface had onLostPointerCapture, we'd call it here on every
                        // object that's getting removed.
                        const { internal: { capturedMap }, } = eventsStateGetter();
                        capturedMap.delete(event.pointerId);
                    }
                    cancelPointer([]);
                };
        }
        // Any other pointer goes here ...
        return (event) => {
            const { pointermissed, internal } = eventsStateGetter();
            prepareRay(event);
            // Get fresh intersects
            const isPointerMove = name === 'pointermove';
            const isClickEvent = name === 'click' || name === 'contextmenu' || name === 'dblclick';
            const filter = isPointerMove ? filterPointerEvents : undefined;
            const hits = patchIntersects(intersect(filter), event);
            const delta = isClickEvent ? calculateDistance(event) : 0;
            // Save initial coordinates on pointer-down
            if (name === 'pointerdown') {
                internal.initialClick = [event.offsetX, event.offsetY];
                internal.initialHits = hits.map((hit) => hit.eventObject);
            }
            // If a click yields no results, pass it back to the user as a miss
            // Missed events have to come first in order to establish user-land side-effect clean up
            if (isClickEvent && !hits.length) {
                if (delta <= 2) {
                    pointerMissed(event, internal.interaction);
                    if (pointermissed)
                        pointermissed(event);
                }
            }
            // Take care of unhover
            if (isPointerMove)
                cancelPointer(hits);
            handleIntersects(hits, event, delta, (data) => {
                var _a, _b, _c;
                const eventObject = data.eventObject;
                const instance = eventObject.__ngt;
                const handlers = instance === null || instance === void 0 ? void 0 : instance.handlers;
                // Check presence of handlers
                if (!(instance === null || instance === void 0 ? void 0 : instance.eventCount))
                    return;
                if (isPointerMove) {
                    // Move event ...
                    if ((handlers === null || handlers === void 0 ? void 0 : handlers.pointerover) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerenter) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerout) ||
                        (handlers === null || handlers === void 0 ? void 0 : handlers.pointerleave)) {
                        // When enter or out is present take care of hover-state
                        const id = makeId(data);
                        const hoveredItem = internal.hovered.get(id);
                        if (!hoveredItem) {
                            // If the object wasn't previously hovered, book it and call its handler
                            internal.hovered.set(id, data);
                            (_a = handlers.pointerover) === null || _a === void 0 ? void 0 : _a.call(handlers, data);
                            (_b = handlers.pointerenter) === null || _b === void 0 ? void 0 : _b.call(handlers, data);
                        }
                        else if (hoveredItem.stopped) {
                            // If the object was previously hovered and stopped, we shouldn't allow other items to proceed
                            data.stopPropagation();
                        }
                    }
                    // Call mouse move
                    (_c = handlers === null || handlers === void 0 ? void 0 : handlers.pointermove) === null || _c === void 0 ? void 0 : _c.call(handlers, data);
                }
                else {
                    // All other events ...
                    const handler = handlers === null || handlers === void 0 ? void 0 : handlers[name];
                    if (handler) {
                        // Forward all events back to their respective handlers with the exception of click events,
                        // which must use the initial target
                        if (!isClickEvent || internal.initialHits.includes(eventObject)) {
                            // Missed events have to come first
                            pointerMissed(event, internal.interaction.filter((object) => !internal.initialHits.includes(object)));
                            // Now call the handler
                            handler(data);
                        }
                    }
                }
            });
        };
    };
    function pointerMissed(event, objects) {
        objects.forEach((object) => {
            var _a, _b, _c;
            return (_c = (_b = (_a = object.__ngt) === null || _a === void 0 ? void 0 : _a.handlers) === null || _b === void 0 ? void 0 : _b.pointermissed) === null || _c === void 0 ? void 0 : _c.call(_b, event);
        });
    }
    return { handlePointer };
}

const names = {
    click: false,
    contextmenu: false,
    dblclick: false,
    wheel: true,
    pointerdown: true,
    pointerup: true,
    pointerleave: true,
    pointermove: true,
    pointercancel: true,
    lostpointercapture: true,
};
class EventsStore extends EnhancedComponentStore {
    constructor(canvasStore) {
        super({
            connected: false,
            handlers: {},
            internal: {
                interaction: [],
                hovered: new Map(),
                capturedMap: new Map(),
                initialClick: [0, 0],
                initialHits: [],
            },
        });
        this.canvasStore = canvasStore;
        this.initHandlersEffect = this.effect(($) => $.pipe(tap(() => {
            const { handlePointer } = createEvents(() => this.canvasStore.getImperativeState(), () => this.getImperativeState());
            this.patchState({
                handlers: Object.keys(names).reduce((handlers, supportedEventName) => {
                    handlers[supportedEventName] = handlePointer(supportedEventName);
                    return handlers;
                }, {}),
            });
        })));
        this.addInteraction = this.updater((state, interaction) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: [...state.internal.interaction, interaction] }) })));
        this.removeInteraction = this.updater((state, uuid) => (Object.assign(Object.assign({}, state), { internal: Object.assign(Object.assign({}, state.internal), { interaction: state.internal.interaction.filter((obj) => obj.uuid !== uuid) }) })));
        this.connectEffect = this.effect((element$) => element$.pipe(withLatestFrom(this.selectors.handlers$), tap(([element, handlers]) => {
            this.disconnectEffect();
            this.patchState({ connected: element });
            Object.entries(handlers !== null && handlers !== void 0 ? handlers : []).forEach(([name, event]) => {
                const passive = names[name];
                element.addEventListener(name, event, { passive });
            });
        })));
        this.disconnectEffect = this.effect(($) => $.pipe(withLatestFrom(this.selectors.connected$, this.selectors.handlers$), tap(([, connected, handlers]) => {
            if (connected) {
                Object.entries(handlers !== null && handlers !== void 0 ? handlers : []).forEach(([name, event]) => {
                    if (connected instanceof HTMLElement) {
                        connected.removeEventListener(name, event);
                    }
                });
                this.patchState({ connected: false });
            }
        })));
        this.initHandlersEffect();
    }
    ngOnDestroy() {
        this.disconnectEffect();
        super.ngOnDestroy();
    }
}
EventsStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: EventsStore, deps: [{ token: CanvasStore }], target: i0.ɵɵFactoryTarget.Injectable });
EventsStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: EventsStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: EventsStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: CanvasStore }]; } });

class InstancesStore extends EnhancedComponentStore {
    constructor() {
        super({ materials: {}, geometries: {}, objects: {} });
        this.saveMaterial = this.updater((state, { material, id = material.uuid }) => (Object.assign(Object.assign({}, state), { materials: Object.assign(Object.assign({}, state.materials), { [id]: material }) })));
        this.removeMaterial = this.updater((state, id) => {
            const _a = state.materials, _b = id, _ = _a[_b], materials = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { materials });
        });
        this.saveGeometry = this.updater((state, { geometry, id = geometry.uuid }) => (Object.assign(Object.assign({}, state), { bufferGeometries: Object.assign(Object.assign({}, state.geometries), { [id]: geometry }) })));
        this.removeGeometry = this.updater((state, id) => {
            const _a = state.geometries, _b = id, _ = _a[_b], geometries = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { geometries });
        });
        this.saveObject = this.updater((state, obj) => (Object.assign(Object.assign({}, state), { objects: Object.assign(Object.assign({}, state.objects), { [obj.uuid]: obj }) })));
        this.removeObject = this.updater((state, id) => {
            const _a = state.objects, _b = id, _ = _a[_b], objects = __rest(_a, [typeof _b === "symbol" ? _b : _b + ""]);
            return Object.assign(Object.assign({}, state), { objects });
        });
    }
}
InstancesStore.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: InstancesStore, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
InstancesStore.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: InstancesStore });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: InstancesStore, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return []; } });

class NgtCanvasComponent {
    constructor(canvasStore, animationStore, eventsStore, loopService, ngZone, hostElement, document, destroyed) {
        this.canvasStore = canvasStore;
        this.animationStore = animationStore;
        this.eventsStore = eventsStore;
        this.loopService = loopService;
        this.ngZone = ngZone;
        this.hostElement = hostElement;
        this.document = document;
        this.destroyed = destroyed;
        this.hostClass = true;
        this.raycaster = {};
        this.created = new EventEmitter();
    }
    set orthographic(v) {
        this.canvasStore.setIsOrthographic(v);
    }
    set linear(v) {
        this.canvasStore.setIsLinear(v);
    }
    set shadows(v) {
        this.canvasStore.setShadows(v);
    }
    set alpha(v) {
        this.canvasStore.setAlpha(v);
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            var _a;
            this.canvasStore.setSize({
                width: this.hostElement.nativeElement.clientWidth,
                height: this.hostElement.nativeElement.clientHeight,
            });
            this.canvasStore.setDpr(((_a = this.document.defaultView) === null || _a === void 0 ? void 0 : _a.devicePixelRatio) || 1);
            this.canvasStore.initRendererEffect(this.rendererCanvas.nativeElement);
            this.canvasStore.initSceneEffect(this.scene);
            this.canvasStore.initCameraEffect(this.camera);
            this.canvasStore.initRaycasterEffect(this.raycaster);
            this.initWindowResizeListener();
            this.initActiveListener();
        });
    }
    initWindowResizeListener() {
        var _a;
        if ((_a = this.document) === null || _a === void 0 ? void 0 : _a.defaultView) {
            fromEvent(this.document.defaultView, 'resize')
                .pipe(takeUntil(this.destroyed))
                .subscribe(() => {
                this.ngZone.runOutsideAngular(() => {
                    var _a;
                    this.canvasStore.windowResizeEffect({
                        size: {
                            width: this.hostElement.nativeElement.clientWidth,
                            height: this.hostElement.nativeElement.clientHeight,
                        },
                        dpr: ((_a = this.document.defaultView) === null || _a === void 0 ? void 0 : _a.devicePixelRatio) || 1,
                    });
                });
            });
        }
    }
    initActiveListener() {
        this.canvasStore.selectors.internal$
            .pipe(takeUntil(this.destroyed), observeOn(asapScheduler))
            .subscribe(({ active }) => {
            this.ngZone.runOutsideAngular(() => {
                if (active) {
                    const { renderer, camera, scene } = this.canvasStore.getImperativeState();
                    if (renderer && camera && scene) {
                        this.created.emit({ gl: renderer, camera, scene });
                        this.eventsStore.connectEffect(renderer.domElement);
                        this.loopService.start();
                    }
                }
            });
        });
    }
}
NgtCanvasComponent.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCanvasComponent, deps: [{ token: CanvasStore, self: true }, { token: AnimationStore, self: true }, { token: EventsStore, self: true }, { token: LoopService, self: true }, { token: i0.NgZone }, { token: i0.ElementRef }, { token: DOCUMENT, optional: true }, { token: DestroyedService }], target: i0.ɵɵFactoryTarget.Component });
NgtCanvasComponent.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "12.0.0", version: "12.2.13", type: NgtCanvasComponent, selector: "ngt-canvas", inputs: { orthographic: "orthographic", linear: "linear", shadows: "shadows", alpha: "alpha", camera: "camera", scene: "scene", raycaster: "raycaster" }, outputs: { created: "created" }, host: { properties: { "class.ngt-canvas": "this.hostClass" } }, providers: [
        CanvasStore,
        EventsStore,
        InstancesStore,
        AnimationStore,
        LoopService,
        DestroyedService,
    ], viewQueries: [{ propertyName: "rendererCanvas", first: true, predicate: ["rendererCanvas"], descendants: true, static: true }], exportAs: ["ngtCanvas"], ngImport: i0, template: ` <canvas #rendererCanvas></canvas> `, isInline: true, styles: ["\n      :host {\n        display: block;\n        position: relative;\n        width: 100%;\n        height: 100%;\n        overflow: hidden;\n      }\n\n      :host canvas {\n        display: block;\n      }\n    "], changeDetection: i0.ChangeDetectionStrategy.OnPush });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCanvasComponent, decorators: [{
            type: Component,
            args: [{
                    selector: 'ngt-canvas',
                    exportAs: 'ngtCanvas',
                    template: ` <canvas #rendererCanvas></canvas> `,
                    styles: [
                        `
      :host {
        display: block;
        position: relative;
        width: 100%;
        height: 100%;
        overflow: hidden;
      }

      :host canvas {
        display: block;
      }
    `,
                    ],
                    changeDetection: ChangeDetectionStrategy.OnPush,
                    providers: [
                        CanvasStore,
                        EventsStore,
                        InstancesStore,
                        AnimationStore,
                        LoopService,
                        DestroyedService,
                    ],
                }]
        }], ctorParameters: function () { return [{ type: CanvasStore, decorators: [{
                    type: Self
                }] }, { type: AnimationStore, decorators: [{
                    type: Self
                }] }, { type: EventsStore, decorators: [{
                    type: Self
                }] }, { type: LoopService, decorators: [{
                    type: Self
                }] }, { type: i0.NgZone }, { type: i0.ElementRef }, { type: Document, decorators: [{
                    type: Optional
                }, {
                    type: Inject,
                    args: [DOCUMENT]
                }] }, { type: DestroyedService }]; }, propDecorators: { hostClass: [{
                type: HostBinding,
                args: ['class.ngt-canvas']
            }], orthographic: [{
                type: Input
            }], linear: [{
                type: Input
            }], shadows: [{
                type: Input
            }], alpha: [{
                type: Input
            }], camera: [{
                type: Input
            }], scene: [{
                type: Input
            }], raycaster: [{
                type: Input
            }], created: [{
                type: Output
            }], rendererCanvas: [{
                type: ViewChild,
                args: ['rendererCanvas', { static: true }]
            }] } });

// GENERATED
class NgtObject3dController {
    constructor() {
        this.castShadow = false;
        this.receiveShadow = false;
        this.visible = true;
        this.matrixAutoUpdate = true;
        this.appendMode = 'immediate';
        // events
        this.click = new EventEmitter();
        this.contextmenu = new EventEmitter();
        this.dblclick = new EventEmitter();
        this.pointerup = new EventEmitter();
        this.pointerdown = new EventEmitter();
        this.pointerover = new EventEmitter();
        this.pointerout = new EventEmitter();
        this.pointerenter = new EventEmitter();
        this.pointerleave = new EventEmitter();
        this.pointermove = new EventEmitter();
        this.pointermissed = new EventEmitter();
        this.pointercancel = new EventEmitter();
        this.wheel = new EventEmitter();
        this.change$ = new ReplaySubject(1);
    }
    ngOnChanges(changes) {
        this.change$.next(changes);
    }
    get object3dProps() {
        return {
            name: this.name,
            position: this.position,
            rotation: this.rotation,
            quaternion: this.quaternion,
            scale: this.scale,
            color: this.color,
            userData: this.userData,
            dispose: this.dispose,
            castShadow: this.castShadow,
            receiveShadow: this.receiveShadow,
            visible: this.visible,
            matrixAutoUpdate: this.matrixAutoUpdate,
        };
    }
}
NgtObject3dController.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3dController, deps: [], target: i0.ɵɵFactoryTarget.Directive });
NgtObject3dController.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtObject3dController, selector: "\n    ngt-primitive,\n    ngt-mesh,\n    ngt-instanced-mesh,\n    ngt-skinned-mesh,\n    ngt-bone,\n    ngt-line2,\n    ngt-line-segments2,\n    ngt-wireframe,\n    ngt-group,\n    ngt-lod,\n    ngt-scene,\n    ngt-points,\n    ngt-cube-camera,\n    ngt-contact-shadows,\n    ngt-html,\n    ngt-light-probe,\n    ngt-ambient-light,\n    ngt-ambient-light-probe,\n    ngt-hemisphere-light,\n    ngt-hemisphere-light-probe,\n    ngt-directional-light,\n    ngt-point-light,\n    ngt-spot-light,\n    ngt-rect-area-light,\n    ngt-arrow-helper,\n    ngt-axes-helper,\n    ngt-box-helper,\n    ngt-box3-helper,\n    ngt-grid-helper,\n    ngt-camera-helper,\n    ngt-directional-light-helper,\n    ngt-hemisphere-light-helper,\n    ngt-plane-helper,\n    ngt-point-light-helper,\n    ngt-polar-grid-helper,\n    ngt-skeleton-helper,\n    ngt-spot-light-helper,\n    ngt-light-probe-helper,\n    ngt-positional-audio-helper,\n    ngt-rect-area-light-helper,\n    ngt-vertex-normals-helper,\n    ngt-vertex-tangents-helper,\n    ngt-line,\n    ngt-line-loop,\n    ngt-line-segments,\n    ngt-sprite,\n    ngt-camera,\n    ngt-perspective-camera,\n    ngt-orthographic-camera,\n    ngt-array-camera,\n    ngt-stereo-camera\n  ", inputs: { name: "name", position: "position", rotation: "rotation", quaternion: "quaternion", scale: "scale", color: "color", userData: "userData", dispose: "dispose", castShadow: "castShadow", receiveShadow: "receiveShadow", visible: "visible", matrixAutoUpdate: "matrixAutoUpdate", appendMode: "appendMode", appendTo: "appendTo" }, outputs: { click: "click", contextmenu: "contextmenu", dblclick: "dblclick", pointerup: "pointerup", pointerdown: "pointerdown", pointerover: "pointerover", pointerout: "pointerout", pointerenter: "pointerenter", pointerleave: "pointerleave", pointermove: "pointermove", pointermissed: "pointermissed", pointercancel: "pointercancel", wheel: "wheel" }, exportAs: ["ngtObject3dController"], usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3dController, decorators: [{
            type: Directive,
            args: [{
                    selector: `
    ngt-primitive,
    ngt-mesh,
    ngt-instanced-mesh,
    ngt-skinned-mesh,
    ngt-bone,
    ngt-line2,
    ngt-line-segments2,
    ngt-wireframe,
    ngt-group,
    ngt-lod,
    ngt-scene,
    ngt-points,
    ngt-cube-camera,
    ngt-contact-shadows,
    ngt-html,
    ngt-light-probe,
    ngt-ambient-light,
    ngt-ambient-light-probe,
    ngt-hemisphere-light,
    ngt-hemisphere-light-probe,
    ngt-directional-light,
    ngt-point-light,
    ngt-spot-light,
    ngt-rect-area-light,
    ngt-arrow-helper,
    ngt-axes-helper,
    ngt-box-helper,
    ngt-box3-helper,
    ngt-grid-helper,
    ngt-camera-helper,
    ngt-directional-light-helper,
    ngt-hemisphere-light-helper,
    ngt-plane-helper,
    ngt-point-light-helper,
    ngt-polar-grid-helper,
    ngt-skeleton-helper,
    ngt-spot-light-helper,
    ngt-light-probe-helper,
    ngt-positional-audio-helper,
    ngt-rect-area-light-helper,
    ngt-vertex-normals-helper,
    ngt-vertex-tangents-helper,
    ngt-line,
    ngt-line-loop,
    ngt-line-segments,
    ngt-sprite,
    ngt-camera,
    ngt-perspective-camera,
    ngt-orthographic-camera,
    ngt-array-camera,
    ngt-stereo-camera
  `,
                    exportAs: 'ngtObject3dController',
                }]
        }], propDecorators: { name: [{
                type: Input
            }], position: [{
                type: Input
            }], rotation: [{
                type: Input
            }], quaternion: [{
                type: Input
            }], scale: [{
                type: Input
            }], color: [{
                type: Input
            }], userData: [{
                type: Input
            }], dispose: [{
                type: Input
            }], castShadow: [{
                type: Input
            }], receiveShadow: [{
                type: Input
            }], visible: [{
                type: Input
            }], matrixAutoUpdate: [{
                type: Input
            }], appendMode: [{
                type: Input
            }], appendTo: [{
                type: Input
            }], click: [{
                type: Output
            }], contextmenu: [{
                type: Output
            }], dblclick: [{
                type: Output
            }], pointerup: [{
                type: Output
            }], pointerdown: [{
                type: Output
            }], pointerover: [{
                type: Output
            }], pointerout: [{
                type: Output
            }], pointerenter: [{
                type: Output
            }], pointerleave: [{
                type: Output
            }], pointermove: [{
                type: Output
            }], pointermissed: [{
                type: Output
            }], pointercancel: [{
                type: Output
            }], wheel: [{
                type: Output
            }] } });

class NgtCoreModule {
}
NgtCoreModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCoreModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtCoreModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCoreModule, declarations: [NgtCanvasComponent, NgtObject3dController], exports: [NgtCanvasComponent, NgtObject3dController] });
NgtCoreModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCoreModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCoreModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtCanvasComponent, NgtObject3dController],
                    exports: [NgtCanvasComponent, NgtObject3dController],
                }]
        }] });

class NgtAnimationParticipant {
    constructor(animationStore, ngZone) {
        this.animationStore = animationStore;
        this.ngZone = ngZone;
        this.animateReady = new EventEmitter();
    }
    participate(animateObject) {
        this.ngZone.runOutsideAngular(() => {
            if (this.animateReady.observed) {
                if (animateObject instanceof THREE.Object3D) {
                    this.animateTeardown = this.animationStore.registerAnimation(animateObject, (obj, state) => {
                        this.animateReady.emit({
                            animateObject: obj,
                            renderState: state,
                        });
                    });
                }
                else {
                    this.animateTeardown = this.animationStore.registerAnimation((state) => {
                        this.animateReady.emit({
                            animateObject,
                            renderState: state,
                        });
                    });
                }
            }
        });
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.animateTeardown) {
                this.animateTeardown();
            }
        });
    }
}
NgtAnimationParticipant.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtAnimationParticipant, deps: [{ token: AnimationStore }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtAnimationParticipant.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtAnimationParticipant, outputs: { animateReady: "animateReady" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtAnimationParticipant, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: AnimationStore }, { type: i0.NgZone }]; }, propDecorators: { animateReady: [{
                type: Output
            }] } });

const NGT_OBJECT_3D_WATCHED_CONTROLLER = new InjectionToken('Watched Object3D Controller');
const NGT_OBJECT_3D_CONTROLLER_PROVIDER = [
    DestroyedService,
    {
        provide: NGT_OBJECT_3D_WATCHED_CONTROLLER,
        deps: [
            [new Optional(), NgtObject3dController],
            ChangeDetectorRef,
            DestroyedService,
        ],
        useFactory: object3dWatchedControllerFactory,
    },
];
function object3dWatchedControllerFactory(controller, changeDetectorRef, destroy$) {
    if (!controller)
        return null;
    controller.change$.pipe(takeUntil(destroy$)).subscribe(() => {
        changeDetectorRef.markForCheck();
    });
    return controller;
}

const supportedEvents = [
    'click',
    'contextmenu',
    'dblclick',
    'pointerup',
    'pointerdown',
    'pointerover',
    'pointerout',
    'pointerenter',
    'pointerleave',
    'pointermove',
    'pointermissed',
    'pointercancel',
    'wheel',
];
class NgtObject3d extends NgtAnimationParticipant {
    constructor(object3dController, canvasStore, instancesStore, eventsStore, animationStore, ngZone, parentObjectDirective) {
        super(animationStore, ngZone);
        this.object3dController = object3dController;
        this.canvasStore = canvasStore;
        this.instancesStore = instancesStore;
        this.eventsStore = eventsStore;
        this.parentObjectDirective = parentObjectDirective;
        this.$object3d = new BehaviorSubject(null);
        this.object3d$ = this.$object3d.asObservable();
        this.ready = this.object3d$.pipe(filter(Boolean));
        this.changesSubscription = object3dController.change$.subscribe(() => {
            if (this.object3d) {
                this.applyCustomProps();
            }
            if (this.inputChangeHandler) {
                this.inputChangeHandler();
            }
        });
    }
    ngOnChanges() {
        if (this.object3d) {
            this.applyCustomProps();
        }
    }
    init() {
        this.ngZone.runOutsideAngular(() => {
            this.initObject();
            if (this.object3d) {
                this.applyCustomProps();
                const observedEvents = supportedEvents.reduce((result, event) => {
                    if (this.object3dController[event].observed) {
                        result.handlers[event] = this.eventNameToHandler(event);
                        result.eventCount += 1;
                    }
                    return result;
                }, { handlers: {}, eventCount: 0 });
                // add as an interaction if there are events observed
                if (observedEvents.eventCount > 0) {
                    this.eventsStore.addInteraction(this.object3d);
                }
                applyProps(this.object3d, {
                    __ngt: {
                        canvasStateGetter: () => this.canvasStore.getImperativeState(),
                        eventsStateGetter: () => this.eventsStore.getImperativeState(),
                        handlers: observedEvents.handlers,
                        eventCount: observedEvents.eventCount,
                    },
                });
                this.instancesStore.saveObject(this.object3d);
                this.appendToParent();
                this.objectReady();
            }
        });
    }
    addToScene() {
        const { scene } = this.canvasStore.getImperativeState();
        if (scene) {
            scene.add(this.object3d);
        }
    }
    addToParent() {
        if (this.parentObjectDirective) {
            this.parentObjectDirective.object3d.add(this.object3d);
        }
        else {
            this.addToScene();
        }
    }
    remove() {
        if (this.object3dController.appendTo) {
            this.object3dController.appendTo.remove(this.object3d);
        }
        else if (this.parentObjectDirective &&
            this.object3dController.appendMode === 'immediate') {
            this.parentObjectDirective.object3d.remove(this.object3d);
        }
        else {
            const { scene } = this.canvasStore.getImperativeState();
            if (scene) {
                scene.remove(this.object3d);
            }
        }
        this.object3d.clear();
    }
    objectReady() {
        this.$object3d.next(this.object3d);
        this.participate(this.object3d);
    }
    applyCustomProps() {
        this.ngZone.runOutsideAngular(() => {
            const customProps = {
                castShadow: this.object3dController.castShadow,
                receiveShadow: this.object3dController.receiveShadow,
                visible: this.object3dController.visible,
                matrixAutoUpdate: this.object3dController.matrixAutoUpdate,
            };
            if (this.object3dController.name) {
                customProps['name'] = this.object3dController.name;
            }
            if (this.object3dController.position) {
                customProps['position'] = this.object3dController.position;
            }
            if (this.object3dController.rotation) {
                customProps['rotation'] = this.object3dController.rotation;
            }
            else if (this.object3dController.quaternion) {
                customProps['quaternion'] = this.object3dController.quaternion;
            }
            if (this.object3dController.scale) {
                customProps['scale'] = this.object3dController.scale;
            }
            if (this.object3dController.userData) {
                customProps['userData'] = this.object3dController.userData;
            }
            if (this.object3dController.color) {
                this.object3dController.color = Array.isArray(this.object3dController.color)
                    ? new THREE.Color(...this.object3dController.color)
                    : new THREE.Color(this.object3dController.color);
                if (!this.canvasStore.getImperativeState().isLinear) {
                    this.object3dController.color.convertSRGBToLinear();
                }
                customProps['color'] = this.object3dController.color;
            }
            if (this.object3dController.dispose) {
                customProps['dispose'] = this.object3dController.dispose;
            }
            this.object3dController.change$.pipe(take(1)).subscribe((changes) => {
                if (changes) {
                    for (const [inputName, inputChange] of Object.entries(changes)) {
                        if (!inputChange.isFirstChange() ||
                            [
                                'name',
                                'position',
                                'rotation',
                                'quaternion',
                                'scale',
                                'userData',
                                'color',
                                'dispose',
                                'castShadow',
                                'receiveShadow',
                                'visible',
                                'matrixAutoUpdate',
                            ].includes(inputName) // skip 12 common inputs
                        ) {
                            continue;
                        }
                        customProps[inputName] = inputChange.currentValue;
                    }
                }
            });
            applyProps(this.object3d, customProps);
            this.object3d.updateMatrix();
        });
    }
    eventNameToHandler(eventName) {
        return (event) => {
            this.ngZone.run(() => {
                this.object3dController[eventName].emit(event);
            });
        };
    }
    appendToParent() {
        if (this.object3dController.appendTo) {
            this.object3dController.appendTo.add(this.object3d);
            return;
        }
        if (this.object3dController.appendMode === 'root') {
            this.addToScene();
            return;
        }
        if (this.object3dController.appendMode === 'immediate') {
            this.addToParent();
        }
    }
    ngOnDestroy() {
        super.ngOnDestroy();
        if (this.changesSubscription) {
            this.changesSubscription.unsubscribe();
        }
        this.ngZone.runOutsideAngular(() => {
            if (this.object3d) {
                this.remove();
                this.instancesStore.removeObject(this.object3d.uuid);
                this.eventsStore.removeInteraction(this.object3d.uuid);
                this.animationStore.unregisterAnimationEffect(this.object3d.uuid);
            }
        });
    }
}
NgtObject3d.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3d, deps: [{ token: NGT_OBJECT_3D_WATCHED_CONTROLLER }, { token: CanvasStore }, { token: InstancesStore }, { token: EventsStore }, { token: AnimationStore }, { token: i0.NgZone }, { token: NgtObject3d, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtObject3d.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtObject3d, outputs: { ready: "ready" }, usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3d, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: NgtObject3dController, decorators: [{
                    type: Inject,
                    args: [NGT_OBJECT_3D_WATCHED_CONTROLLER]
                }] }, { type: CanvasStore }, { type: InstancesStore }, { type: EventsStore }, { type: AnimationStore }, { type: i0.NgZone }, { type: NgtObject3d, decorators: [{
                    type: Optional
                }, {
                    type: SkipSelf
                }] }]; }, propDecorators: { ready: [{
                type: Output
            }] } });

class NgtPrimitive extends NgtAnimationParticipant {
    constructor(animationStore, ngZone, object3dController, canvasStore, parentObjectDirective) {
        super(animationStore, ngZone);
        this.object3dController = object3dController;
        this.canvasStore = canvasStore;
        this.parentObjectDirective = parentObjectDirective;
        this.ready = new EventEmitter();
    }
    get object() {
        return this._object;
    }
    set object(value) {
        if (value == null) {
            console.error('[object] is required');
        }
        this._object = value;
        if (value) {
            this.ready.emit(value);
            this.participate(value);
        }
    }
    get object3d() {
        return this._object;
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            this.appendToParent();
        });
    }
    ngOnDestroy() {
        super.ngOnDestroy();
        this.ngZone.runOutsideAngular(() => {
            if (this.object3d) {
                this.remove();
            }
        });
    }
    appendToParent() {
        if (this.object3d && this.object3d instanceof THREE.Object3D) {
            if (this.object3dController.appendTo) {
                this.object3dController.appendTo.add(this.object3d);
                return;
            }
            const { scene } = this.canvasStore.getImperativeState();
            if (this.object3dController.appendMode === 'root') {
                if (scene) {
                    scene.add(this.object3d);
                }
                return;
            }
            if (this.object3dController.appendMode === 'immediate') {
                if (this.parentObjectDirective) {
                    this.parentObjectDirective.object3d.add(this.object3d);
                }
                else {
                    if (scene) {
                        scene.add(this.object3d);
                    }
                }
            }
        }
    }
    remove() {
        if (this.object3d instanceof THREE.Object3D) {
            if (this.object3dController.appendTo) {
                this.object3dController.appendTo.remove(this.object3d);
            }
            else if (this.parentObjectDirective &&
                this.object3dController.appendMode === 'immediate') {
                this.parentObjectDirective.object3d.remove(this.object3d);
            }
            else {
                const { scene } = this.canvasStore.getImperativeState();
                if (scene) {
                    scene.remove(this.object3d);
                }
            }
            this.object3d.clear();
        }
    }
}
NgtPrimitive.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitive, deps: [{ token: AnimationStore }, { token: i0.NgZone }, { token: NGT_OBJECT_3D_WATCHED_CONTROLLER }, { token: CanvasStore }, { token: NgtObject3d, optional: true, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtPrimitive.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtPrimitive, selector: "ngt-primitive", inputs: { object: "object" }, outputs: { ready: "ready" }, providers: [
        {
            provide: NgtObject3d,
            useExisting: NgtPrimitive,
        },
        NGT_OBJECT_3D_CONTROLLER_PROVIDER,
    ], exportAs: ["ngtPrimitive"], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitive, decorators: [{
            type: Directive,
            args: [{
                    selector: 'ngt-primitive',
                    exportAs: 'ngtPrimitive',
                    providers: [
                        {
                            provide: NgtObject3d,
                            useExisting: NgtPrimitive,
                        },
                        NGT_OBJECT_3D_CONTROLLER_PROVIDER,
                    ],
                }]
        }], ctorParameters: function () { return [{ type: AnimationStore }, { type: i0.NgZone }, { type: NgtObject3dController, decorators: [{
                    type: Inject,
                    args: [NGT_OBJECT_3D_WATCHED_CONTROLLER]
                }] }, { type: CanvasStore }, { type: NgtObject3d, decorators: [{
                    type: Optional
                }, {
                    type: SkipSelf
                }] }]; }, propDecorators: { ready: [{
                type: Output
            }], object: [{
                type: Input
            }] } });

class NgtPrimitiveModule {
}
NgtPrimitiveModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitiveModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtPrimitiveModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitiveModule, declarations: [NgtPrimitive], exports: [NgtPrimitive] });
NgtPrimitiveModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitiveModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtPrimitiveModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtPrimitive],
                    exports: [NgtPrimitive],
                }]
        }] });

class NgtObject3dMaterialGeometry extends NgtObject3d {
    set extraArgs(v) {
        this._extraArgs = v;
    }
    ngAfterContentInit() {
        this.init();
    }
    initObject() {
        if (this.canCreate()) {
            const material = this.getMaterial();
            const geometry = this.getGeometry();
            this._objectMaterialGeometry = new this.objectType(geometry, material, ...(this._extraArgs || []));
            if (this.morphTargetDictionary &&
                'morphTargetDictionary' in this._objectMaterialGeometry) {
                this._objectMaterialGeometry.morphTargetDictionary =
                    this.morphTargetDictionary;
            }
            if (this.morphTargetInfluences &&
                'morphTargetInfluences' in this._objectMaterialGeometry) {
                this._objectMaterialGeometry.morphTargetInfluences =
                    this.morphTargetInfluences;
            }
            if (this.customize) {
                this.customize();
            }
        }
    }
    canCreate() {
        return true;
    }
    get object3d() {
        return this._objectMaterialGeometry;
    }
    getMaterial() {
        if (this.material) {
            if ((Array.isArray(this.material) &&
                this.material[0] instanceof THREE.Material) ||
                this.material instanceof THREE.Material) {
                return this.material;
            }
            if (Array.isArray(this.material)) {
                return this.material.map((materialId) => this.instancesStore.getImperativeState().materials[materialId]);
            }
            return this.instancesStore.getImperativeState().materials[this.material];
        }
        return undefined;
    }
    getGeometry() {
        if (this.geometry) {
            if (this.geometry instanceof THREE.BufferGeometry) {
                return this.geometry;
            }
            return this.instancesStore.getImperativeState().geometries[this.geometry];
        }
        return undefined;
    }
}
NgtObject3dMaterialGeometry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3dMaterialGeometry, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtObject3dMaterialGeometry.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtObject3dMaterialGeometry, inputs: { geometry: "geometry", material: "material", morphTargetInfluences: "morphTargetInfluences", morphTargetDictionary: "morphTargetDictionary" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtObject3dMaterialGeometry, decorators: [{
            type: Directive
        }], propDecorators: { geometry: [{
                type: Input
            }], material: [{
                type: Input
            }], morphTargetInfluences: [{
                type: Input
            }], morphTargetDictionary: [{
                type: Input
            }] } });

class NgtGeometry {
    constructor(instancesStore, ngZone) {
        this.instancesStore = instancesStore;
        this.ngZone = ngZone;
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            if (!this.geometry) {
                this.init();
            }
        });
    }
    init() {
        this._geometry = new this.geometryType(...this._extraArgs);
        this.instancesStore.saveGeometry({
            id: this.ngtId,
            geometry: this._geometry,
        });
    }
    get geometry() {
        return this._geometry;
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.geometry) {
                this.instancesStore.removeGeometry(this.ngtId || this.geometry.uuid);
                this.geometry.dispose();
            }
        });
    }
}
NgtGeometry.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtGeometry, deps: [{ token: InstancesStore, skipSelf: true }, { token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtGeometry.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtGeometry, inputs: { ngtId: "ngtId" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtGeometry, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: InstancesStore, decorators: [{
                    type: SkipSelf
                }] }, { type: i0.NgZone }]; }, propDecorators: { ngtId: [{
                type: Input
            }] } });

class NgtAttribute {
    constructor(ngZone, geometryDirective) {
        this.ngZone = ngZone;
        this.geometryDirective = geometryDirective;
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnChanges() {
        this.ngZone.runOutsideAngular(() => {
            if (this.attribute) {
                this.attribute.needsUpdate = true;
            }
        });
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            if (!this.attribute) {
                this.init();
            }
        });
    }
    init() {
        if (this.geometryDirective && this.attach) {
            this._attribute = new this.attributeType(...this._extraArgs);
            if (this.attribute) {
                this.geometryDirective.geometry.setAttribute(this.attach, this.attribute);
            }
        }
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.geometryDirective && this.attach) {
                this.geometryDirective.geometry.deleteAttribute(this.attach);
            }
        });
    }
    get attribute() {
        return this._attribute;
    }
}
NgtAttribute.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtAttribute, deps: [{ token: i0.NgZone }, { token: NgtGeometry, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtAttribute.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtAttribute, inputs: { attach: "attach" }, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtAttribute, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: i0.NgZone }, { type: NgtGeometry, decorators: [{
                    type: Optional
                }] }]; }, propDecorators: { attach: [{
                type: Input
            }] } });

class NgtCurve {
    constructor(ngZone, geometryDirective) {
        this.ngZone = ngZone;
        this.geometryDirective = geometryDirective;
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            if (!this.curve) {
                this.init();
            }
        });
    }
    init() {
        this._curve = new this.curveType(...this._extraArgs);
        if (this.curve && this.geometryDirective) {
            const points = this.curve.getPoints(this.divisions);
            this.geometryDirective.geometry.setFromPoints(points);
        }
    }
    get curve() {
        return this._curve;
    }
}
NgtCurve.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCurve, deps: [{ token: i0.NgZone }, { token: NgtGeometry, optional: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtCurve.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtCurve, inputs: { divisions: "divisions" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCurve, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: i0.NgZone }, { type: NgtGeometry, decorators: [{
                    type: Optional
                }] }]; }, propDecorators: { divisions: [{
                type: Input
            }] } });

class NgtCommonCamera extends NgtObject3d {
    constructor() {
        super(...arguments);
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnInit() {
        if (!this.object3d) {
            this.init();
        }
    }
    initObject() {
        this._camera = new this.cameraType(...this._extraArgs);
    }
    get object3d() {
        return this._camera;
    }
}
NgtCommonCamera.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonCamera, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonCamera.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtCommonCamera, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonCamera, decorators: [{
            type: Directive
        }] });

class NgtHelper extends NgtObject3d {
    constructor() {
        super(...arguments);
        this._extraArgs = [];
        this.inputChangeHandler = () => {
            if (!this.object3d) {
                this.init();
            }
        };
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnChanges() {
        super.ngOnChanges();
        this.inputChangeHandler();
    }
    ngOnInit() {
        this.inputChangeHandler();
    }
    initObject() {
        try {
            this._helper = new this.helperType(...this._extraArgs);
        }
        catch (e) {
            console.log('Failed to initialize Helper');
        }
    }
    get object3d() {
        return this._helper;
    }
}
NgtHelper.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtHelper, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtHelper.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtHelper, usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtHelper, decorators: [{
            type: Directive
        }] });

class NgtLight extends NgtObject3d {
    constructor() {
        super(...arguments);
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this.init();
        });
    }
    ngOnInit() {
        if (!this.object3d) {
            this.init();
        }
    }
    initObject() {
        if (this.intensity) {
            this._extraArgs[1] = this.intensity;
        }
        this._light = new this.lightType(...this._extraArgs);
    }
    get object3d() {
        return this._light;
    }
}
NgtLight.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtLight, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtLight.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtLight, inputs: { intensity: "intensity" }, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtLight, decorators: [{
            type: Directive
        }], propDecorators: { intensity: [{
                type: Input
            }] } });

class NgtMaterial {
    constructor(ngZone, instancesStore, canvasStore) {
        this.ngZone = ngZone;
        this.instancesStore = instancesStore;
        this.canvasStore = canvasStore;
    }
    set parameters(v) {
        this._parameters = v;
        if (v && this.material) {
            this.ngZone.runOutsideAngular(() => {
                this.convertColorToLinear(v);
                this.material.setValues(v);
                this.material.needsUpdate = true;
            });
        }
    }
    get parameters() {
        return this._parameters;
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            if (this.parameters) {
                this.convertColorToLinear(this.parameters);
            }
            this._material = new this.materialType(this.parameters);
            this.instancesStore.saveMaterial({
                id: this.ngtId,
                material: this._material,
            });
        });
    }
    get material() {
        return this._material;
    }
    convertColorToLinear(parameters) {
        if ('color' in parameters) {
            const colorParams = parameters['color'];
            parameters['color'] = Array.isArray(colorParams)
                ? new THREE.Color(...colorParams)
                : new THREE.Color(colorParams);
            if (!this.canvasStore.getImperativeState().isLinear) {
                parameters['color'].convertSRGBToLinear();
            }
        }
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.material) {
                this.instancesStore.removeMaterial(this.ngtId || this.material.uuid);
                this.material.dispose();
            }
        });
    }
}
NgtMaterial.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMaterial, deps: [{ token: i0.NgZone }, { token: InstancesStore, skipSelf: true }, { token: CanvasStore, skipSelf: true }], target: i0.ɵɵFactoryTarget.Directive });
NgtMaterial.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtMaterial, inputs: { ngtId: "ngtId", parameters: "parameters" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMaterial, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: i0.NgZone }, { type: InstancesStore, decorators: [{
                    type: SkipSelf
                }] }, { type: CanvasStore, decorators: [{
                    type: SkipSelf
                }] }]; }, propDecorators: { ngtId: [{
                type: Input
            }], parameters: [{
                type: Input
            }] } });

class NgtCommonLine extends NgtObject3dMaterialGeometry {
    set materialDirectives(v) {
        if (this.material == null && v) {
            this.material =
                v.length === 1
                    ? v.first.material
                    : v.toArray().map((dir) => dir.material);
        }
    }
    set bufferGeometryDirective(v) {
        if (this.geometry == null) {
            this.geometry = v.geometry;
        }
    }
    get objectType() {
        return this.lineType;
    }
}
NgtCommonLine.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonLine, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonLine.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtCommonLine, queries: [{ propertyName: "bufferGeometryDirective", first: true, predicate: NgtGeometry, descendants: true }, { propertyName: "materialDirectives", predicate: NgtMaterial }], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonLine, decorators: [{
            type: Directive
        }], propDecorators: { materialDirectives: [{
                type: ContentChildren,
                args: [NgtMaterial]
            }], bufferGeometryDirective: [{
                type: ContentChild,
                args: [NgtGeometry]
            }] } });

class NgtCommonMesh extends NgtObject3dMaterialGeometry {
    set materialDirectives(v) {
        if (this.material == null && v) {
            this.material =
                v.length === 1
                    ? v.first.material
                    : v.toArray().map((dir) => dir.material);
        }
    }
    set bufferGeometryDirective(v) {
        if (this.geometry == null && v) {
            this.geometry = v.geometry;
        }
    }
    get objectType() {
        return this.meshType;
    }
}
NgtCommonMesh.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonMesh, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonMesh.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtCommonMesh, queries: [{ propertyName: "bufferGeometryDirective", first: true, predicate: NgtGeometry, descendants: true }, { propertyName: "materialDirectives", predicate: NgtMaterial }], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonMesh, decorators: [{
            type: Directive
        }], propDecorators: { materialDirectives: [{
                type: ContentChildren,
                args: [NgtMaterial]
            }], bufferGeometryDirective: [{
                type: ContentChild,
                args: [NgtGeometry]
            }] } });

class NgtCommonSprite extends NgtObject3d {
    ngAfterContentInit() {
        this.init();
    }
    initObject() {
        if (this.material) {
            this._sprite = new this.spriteType(this.material);
            return;
        }
        if (this.materialDirective) {
            if (this.materialDirective.material instanceof THREE.SpriteMaterial) {
                this._sprite = new this.spriteType(this.materialDirective.material);
                return;
            }
            console.warn(`Sprite can only be instantiated with SpriteMaterial`);
        }
    }
    get object3d() {
        return this._sprite;
    }
}
NgtCommonSprite.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonSprite, deps: null, target: i0.ɵɵFactoryTarget.Directive });
NgtCommonSprite.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtCommonSprite, inputs: { material: "material" }, queries: [{ propertyName: "materialDirective", first: true, predicate: NgtMaterial, descendants: true }], usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtCommonSprite, decorators: [{
            type: Directive
        }], propDecorators: { material: [{
                type: Input
            }], materialDirective: [{
                type: ContentChild,
                args: [NgtMaterial]
            }] } });

class NgtTexture {
    constructor(ngZone) {
        this.ngZone = ngZone;
        this._extraArgs = [];
    }
    set extraArgs(v) {
        this._extraArgs = v;
        this.ngZone.runOutsideAngular(() => {
            this._texture = new this.textureType(...this._extraArgs);
        });
    }
    ngOnInit() {
        this.ngZone.runOutsideAngular(() => {
            if (!this.texture) {
                this._texture = new this.textureType(...this._extraArgs);
            }
        });
    }
    get texture() {
        return this._texture;
    }
    ngOnDestroy() {
        this.ngZone.runOutsideAngular(() => {
            if (this.texture) {
                this.texture.dispose();
            }
        });
    }
}
NgtTexture.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtTexture, deps: [{ token: i0.NgZone }], target: i0.ɵɵFactoryTarget.Directive });
NgtTexture.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtTexture, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtTexture, decorators: [{
            type: Directive
        }], ctorParameters: function () { return [{ type: i0.NgZone }]; } });

class LoaderService {
    constructor() {
        this.cached = new Map();
    }
    use(loaderConstructor, input, extensions, onProgress) {
        const keys = (Array.isArray(input) ? input : [input]);
        const loader = new loaderConstructor();
        if (extensions) {
            extensions(loader);
        }
        const results$ = forkJoin(keys.map((key) => {
            if (this.cached.has(key)) {
                return of(this.cached.get(key));
            }
            return defer(() => loader.loadAsync(key, onProgress)).pipe(tap((data) => {
                if (data.scene) {
                    Object.assign(data, this.buildGraph(data.scene));
                }
                this.cached.set(key, data);
            }), catchError((err) => {
                console.error(`Error loading ${key}: ${err.message}`);
                return throwError(err);
            }));
        }));
        return defer(() => Array.isArray(input)
            ? results$
            : results$.pipe(map((results) => results[0]))).pipe(shareReplay({ bufferSize: 1, refCount: true }));
    }
    ngOnDestroy() {
        this.cached.clear();
    }
    buildGraph(object) {
        const data = { nodes: {}, materials: {} };
        if (object) {
            object.traverse((obj) => {
                if (obj.name) {
                    data.nodes[obj.name] = obj;
                }
                if (obj.material &&
                    !data.materials[obj
                        .material.name]) {
                    data.materials[obj
                        .material.name] = obj.material;
                }
            });
        }
        return data;
    }
}
LoaderService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoaderService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
LoaderService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoaderService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: LoaderService, decorators: [{
            type: Injectable,
            args: [{
                    providedIn: 'root',
                }]
        }] });

class NgtColorPipe {
    /**
     * ConstructorParameters<typeof THREE.Color> has a limitation on THREE.Color constructor overloads
     */
    transform(args) {
        return new THREE.Color(...args);
    }
}
NgtColorPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtColorPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipe, name: "color" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'color',
                    pure: true,
                }]
        }] });

class NgtColorPipeModule {
}
NgtColorPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtColorPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipeModule, declarations: [NgtColorPipe], exports: [NgtColorPipe] });
NgtColorPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtColorPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtColorPipe],
                    exports: [NgtColorPipe],
                }]
        }] });

class NgtFogPipe {
    transform(args) {
        return new THREE.Fog(...args);
    }
}
NgtFogPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtFogPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipe, name: "fog" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'fog',
                    pure: true,
                }]
        }] });

class NgtFogPipeModule {
}
NgtFogPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtFogPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipeModule, declarations: [NgtFogPipe], exports: [NgtFogPipe] });
NgtFogPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtFogPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtFogPipe],
                    exports: [NgtFogPipe],
                }]
        }] });

class NgtMathPipe {
    transform(value, keyOfMath) {
        const params = Array.isArray(value) ? value : [value];
        return Math[keyOfMath](...params);
    }
}
NgtMathPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtMathPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipe, name: "math" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'math',
                    pure: true,
                }]
        }] });

class NgtMathConstantPipe {
    transform(value, keyOfMathConst) {
        if (keyOfMathConst === 'random') {
            return Math.random();
        }
        return value * Math[keyOfMathConst];
    }
}
NgtMathConstantPipe.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathConstantPipe, deps: [], target: i0.ɵɵFactoryTarget.Pipe });
NgtMathConstantPipe.ɵpipe = i0.ɵɵngDeclarePipe({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathConstantPipe, name: "mathConst" });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathConstantPipe, decorators: [{
            type: Pipe,
            args: [{
                    name: 'mathConst',
                    pure: true,
                }]
        }] });

class NgtMathPipeModule {
}
NgtMathPipeModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipeModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtMathPipeModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipeModule, declarations: [NgtMathPipe, NgtMathConstantPipe], exports: [NgtMathPipe, NgtMathConstantPipe] });
NgtMathPipeModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipeModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtMathPipeModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtMathPipe, NgtMathConstantPipe],
                    exports: [NgtMathPipe, NgtMathConstantPipe],
                }]
        }] });

const MAX_VALUE = 0x10000;
class NgtRepeatDirective {
    constructor(viewContainer, templateRef) {
        this.viewContainer = viewContainer;
        this.templateRef = templateRef;
    }
    set repeatOf(count) {
        const safeCount = Math.floor(Math.max(0, Math.min(count, MAX_VALUE)));
        const { length } = this.viewContainer;
        if (safeCount < length) {
            this.removeContainers(length - safeCount);
        }
        else {
            this.addContainers(length, safeCount);
        }
    }
    addContainers(length, count) {
        for (let index = length; index < count; index++) {
            this.viewContainer.createEmbeddedView(this.templateRef, {
                $implicit: index,
                isFirst: index === length,
                isLast: index === count - 1,
                isOdd: !(index % 2),
                isEven: !!(index % 2),
            });
        }
    }
    removeContainers(amount) {
        for (let index = 0; index < amount; index++) {
            this.viewContainer.remove();
        }
    }
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
}
NgtRepeatDirective.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatDirective, deps: [{ token: ViewContainerRef }, { token: TemplateRef }], target: i0.ɵɵFactoryTarget.Directive });
NgtRepeatDirective.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "12.0.0", version: "12.2.13", type: NgtRepeatDirective, selector: "[repeat][repeatOf]", inputs: { repeatOf: "repeatOf" }, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatDirective, decorators: [{
            type: Directive,
            args: [{
                    selector: '[repeat][repeatOf]',
                }]
        }], ctorParameters: function () { return [{ type: i0.ViewContainerRef, decorators: [{
                    type: Inject,
                    args: [ViewContainerRef]
                }] }, { type: i0.TemplateRef, decorators: [{
                    type: Inject,
                    args: [TemplateRef]
                }] }]; }, propDecorators: { repeatOf: [{
                type: Input
            }] } });

class NgtRepeatModule {
}
NgtRepeatModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
NgtRepeatModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatModule, declarations: [NgtRepeatDirective], exports: [NgtRepeatDirective] });
NgtRepeatModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatModule });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.2.13", ngImport: i0, type: NgtRepeatModule, decorators: [{
            type: NgModule,
            args: [{
                    declarations: [NgtRepeatDirective],
                    exports: [NgtRepeatDirective],
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { AnimationStore, CanvasStore, DestroyedService, EnhancedComponentStore, EventsStore, InstancesStore, LoaderService, LoopService, NGT_OBJECT_3D_CONTROLLER_PROVIDER, NGT_OBJECT_3D_WATCHED_CONTROLLER, NgtAnimationParticipant, NgtAttribute, NgtCanvasComponent, NgtColorPipe, NgtColorPipeModule, NgtCommonCamera, NgtCommonLine, NgtCommonMesh, NgtCommonSprite, NgtCoreModule, NgtCurve, NgtFogPipe, NgtFogPipeModule, NgtGeometry, NgtHelper, NgtLight, NgtMaterial, NgtMathConstantPipe, NgtMathPipe, NgtMathPipeModule, NgtObject3d, NgtObject3dController, NgtObject3dMaterialGeometry, NgtPrimitive, NgtPrimitiveModule, NgtRepeatDirective, NgtRepeatModule, NgtTexture, applyProps, getSelectors, isOrthographicCamera, object3dWatchedControllerFactory };
//# sourceMappingURL=angular-three-core.js.map
